@@ -3,6 +3,7 @@ import traceback
 from pathlib import Path
 import multiprocessing as mp
 from functools import partial
+import time
 
 import numpy as np
 import pandas as pd
@@ -12,6 +13,7 @@ from lightwood.constants.lightwood import ColumnDataTypes
 from mindsdb_native.libs.constants.mindsdb import *
 from mindsdb_native.config import *
 from mindsdb_native.libs.helpers.general_helpers import evaluate_accuracy
+from mindsdb_native.libs.helpers.mp_helpers import get_nr_procs
 
 
 def _make_pred(row):
@@ -127,13 +129,12 @@ class LightwoodBackend:
         if len(gb_arr) > 0:
             df_arr = []
             for _, df in original_df.groupby(gb_arr):
-                df.sort_values(by=ob_arr, inplace=True)
-                df_arr.append(df)
+                df_arr.append(df.sort_values(by=ob_arr))
         else:
             df_arr = [original_df]
 
         if len(original_df) > 500:
-            pool = mp.Pool(processes = (mp.cpu_count() - 1))
+            pool = mp.Pool(processes=get_nr_procs())
             # Make type `object` so that dataframe cells can be python lists
             df_arr = pool.map(partial(_ts_to_obj, historical_columns=ob_arr + self.transaction.lmd['tss']['historical_columns']), df_arr)
             df_arr = pool.map(partial(_ts_order_col_to_cell_lists, historical_columns=ob_arr + self.transaction.lmd['tss']['historical_columns']), df_arr)
@@ -191,7 +192,7 @@ class LightwoodBackend:
         config['input_features'] = []
         config['output_features'] = []
 
-        for col_name in self.transaction.input_data.columns:
+        for col_name in self.transaction.lmd['columns']:
             if col_name in self.transaction.lmd['columns_to_ignore'] or col_name not in self.transaction.lmd['stats_v2']:
                 continue
 
@@ -245,15 +246,16 @@ class LightwoodBackend:
                 lightwood_data_type = ColumnDataTypes.TIME_SERIES
                 self.nn_mixer_only = True
 
+            grouped_by = self.transaction.lmd['tss'].get('group_by', [])
             col_config = {
                 'name': col_name,
-                'type': lightwood_data_type
+                'type': lightwood_data_type,
+                'grouped_by': col_name in grouped_by if grouped_by else False
             }
 
             if data_subtype == DATA_SUBTYPES.SHORT:
                 col_config['encoder_class'] = lightwood.encoders.text.short.ShortTextEncoder
 
-
             if col_name in self.transaction.lmd['weight_map']:
                 col_config['weights'] = self.transaction.lmd['weight_map'][col_name]
 
@@ -268,7 +270,8 @@ class LightwoodBackend:
 
                 if self.transaction.lmd['tss']['is_timeseries']:
                     col_config['additional_info'] = {
-                        'nr_predictions': self.transaction.lmd['tss']['nr_predictions']
+                        'nr_predictions': self.transaction.lmd['tss']['nr_predictions'],
+                        'time_series_target': True,
                     }
                 config['output_features'].append(col_config)
 
@@ -380,19 +383,33 @@ class LightwoodBackend:
                     for mx_cls in lightwood.mixers.BaseMixer.__subclasses__():
                         if mx_cls.__name__ == mixer_class:
                             mixer_class = mx_cls
+                            break
                     else:
-                        raise ValueError(f'Mixer "{mixer_classes[i]}" doesn\'t exist')
+                        raise ValueError(f'Mixer "{mixer_class}" doesn\'t exist')
                 if mixer_class is not None:
                     final_mixer_classes.append(mixer_class)
 
             if len(final_mixer_classes) == 0:
                 raise Exception(f'No valid mixers')
 
+            # @TODO Might be better to turn this into a function
             stop_training_after = self.transaction.lmd['stop_training_in_x_seconds']
             if stop_training_after is None:
                 # Stop training after 12 hours unless the user doesn't want us to
                 stop_training_after = 3600 * 12
-            stop_training_after_per_mixer = stop_training_after/len(mixer_classes)
+
+            took_thus_far = int(time.time() - self.transaction.lmd['learn_started_at'])
+            remaining_time = int(max(0, stop_training_after - took_thus_far))
+            if took_thus_far*2 > stop_training_after:
+                new_remaining_time = stop_training_after
+                self.transaction.log.warning(f'You asked your predictor to stop training too quickly, the data preparation phase alone took {took_thus_far}. We\'d be left with only {remaining_time} to train the underlying machine learning models and analyze them. Will instead try to spend {new_remaining_time} seconds doing so')
+                remaining_time = new_remaining_time
+
+            # We don't know how long the model analysis will last, so let's allocate 1/4th of the remaining time for training to it
+            data_analysis_takes = 1/4
+
+            training_time_per_mixer = (remaining_time*(1-data_analysis_takes))/len(mixer_classes)
+            # @TODO Might be better to turn this into a function
 
             for mixer_class in final_mixer_classes:
                 lightwood_config['mixer']['kwargs'] = {}
@@ -410,7 +427,7 @@ class LightwoodBackend:
                     lightwood_config['mixer']['kwargs']['callback_on_iter'] = self.callback_on_iter
                     lightwood_config['mixer']['kwargs']['eval_every_x_epochs'] = eval_every_x_epochs / len(final_mixer_classes)
 
-                lightwood_config['mixer']['kwargs']['stop_training_after_seconds'] = stop_training_after_per_mixer
+                lightwood_config['mixer']['kwargs']['stop_training_after_seconds'] = training_time_per_mixer
 
                 self.predictor = lightwood.Predictor(lightwood_config.copy())
 
